(******************************************************************************)
(** * Adding one event to en execution graph        *)
(** *  and picking w_max as the previous write      *)
(******************************************************************************)

Require Import Omega.

Require Import Hahn.
Require Import AuxRel.

Require Import Labels Events Execution FG.
Require Import Wmax.
Set Implicit Arguments.
Remove Hints plus_n_O.

Notation "'R'" := is_r.
Notation "'W'" := is_w.
Notation "'RMW'" := is_rmw.

Section ADD_MAX.

Lemma w_max_init x : 
  w_max Ginit x ≡₁ is_init ∩₁ (fun w => loc w = x).
Proof.
unfold w_max in *.
unfold Ginit, acts_set, is_init; simpl.
unfolder; split; ins; desf.
apply in_map_iff in H; desf; eauto.
rewrite in_map_iff.
splits; eauto.
exists x; splits; eauto.
apply Loc_list_prop.
Qed.

Lemma w_max_add_read G G' t n w x l w' 
  (READ: is_r_l l) (NRMW: ~ is_rmw_l l) (MAX: w_max G x w')
  (ADD: add G G' t n l w) : w_max G' x w'.
Proof.
unfold add, w_max, max_elt in *; desc.
unfolder; unfolder in MAX; desc; splits; try done.
by apply ACTS; left.
ins; eapply MAX0.
apply CO2; try edone. type_solver.
Qed.

Lemma w_max_add_read_inv G G' t n w x l w' 
  (READ: is_r_l l) (NRMW: ~ is_rmw_l l) (MAX: w_max G' x w')
  (ADD: add G G' t n l w) : w_max G x w'.
Proof.
unfold add, w_max, max_elt in *; desc.
unfolder; unfolder in MAX; desc; splits; try done.
- unfolder in ACTS; desc.
specialize (ACTS w' MAX); desf.
type_solver.
- ins; eapply MAX0.
apply CO2; try edone. type_solver.
Qed.

Lemma w_max_add_read' G G' t n w x l
  (READ: is_r_l l) (NRMW: ~ is_rmw_l l)
  (ADD: add G G' t n l w) : w_max G' x  ≡₁ w_max G x.
Proof.
unfolder; split; ins.
eby eapply w_max_add_read_inv.
eby eapply w_max_add_read.
Qed.

Lemma w_max_add_neq_loc G (WF: Wf G) G' t n w x l w'
  (MAX: w_max G x w') (ADD: add G G' t n l w) (NEQ: loc w <> loc w') : 
  w_max G' x w'.
Proof.
unfold add, w_max, max_elt in *; desc.
unfolder; unfolder in MAX; desc; splits; try done.
by apply ACTS; left.
ins; eapply MAX0.
destruct l; ins.
- eby apply CO2.
- unfold add_co in *.
apply CO1 in REL; [|auto].
unfolder in REL; desf.
apply WF.(wf_col) in REL; unfold same_loc in *.
exfalso; eauto.
- unfold add_co in *.
apply CO1 in REL; [|auto].
unfolder in REL; desf.
apply WF.(wf_col) in REL; unfold same_loc in *.
exfalso; eauto.
Qed.

Lemma w_max_add_neq_loc_inv G G' t n w x l w' 
  (NEQ: loc w <> loc w') (MAX: w_max G' x w')
  (ADD: add G G' t n l w) : w_max G x w'.
Proof.
destruct (is_w_l l) eqn:A.
2: by eapply w_max_add_read_inv; try edone; try type_solver.
unfold add, w_max, max_elt in *.
unfolder; unfolder in MAX; desc; splits; try done.
{ unfolder in ACTS; desc.
specialize (ACTS w' MAX); destruct ACTS; eauto; subst. 
unfold loc in *; ins. }
ins; eapply MAX0.
apply CO1 in A; hahn_rewrite A. 
unfold add_co; basic_solver.
Qed.

Lemma w_max_add_neq_loc' G G' t n w x l (WF: Wf G) 
  (NEQ: loc w <> x) (ADD: add G G' t n l w) : 
  w_max G' x  ≡₁ w_max G x.
Proof.
unfolder; split; ins.
assert (loc x0 = x); subst.
by unfold w_max in *; unfolder in H; desf.
eapply w_max_add_neq_loc_inv; try edone.
assert (loc x0 = x); subst.
by unfold w_max in *; unfolder in H; desf.
eapply w_max_add_neq_loc; try edone.
Qed.

Lemma w_max_add_write G (WF: Wf G) G' t n x l 
  (WRITE: is_w_l l) w (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  w_max G' x (ThreadEvent t n l).
Proof.
unfold add, w_max, max_elt in *.
unfolder; unfolder in MAX; desc; splits; try done.
apply ACTS; basic_solver.
by unfold loc, loc_l in *; ins; desf.
ins.
apply CO1 in REL; [|type_solver].
unfold add_co in REL.
unfolder in REL; desf.
- apply (dom_l (wf_coE WF)) in REL; unfolder in REL; desc.
unfold acts_set in REL.
assert (index (ThreadEvent t n l) < n) by eauto.
unfold index in *; desf; omega.
- 
assert (index (ThreadEvent t n l) < n) by eauto.
unfold index in *; desf; omega.
- apply (dom_l (wf_coE WF)) in REL; unfolder in REL; desc.
unfold acts_set in REL.
assert (index (ThreadEvent t n l) < n) by eauto.
unfold index in *; desf; omega.
- eby eapply MAX0.
Qed.

Lemma w_max_add_write_inv G (WF: Wf G) G' t n x l 
  (WRITE: is_w_l l) w (MAX: w_max G' x (ThreadEvent t n l)) 
  (ADD: add G G' t n l w) : w_max G x w.
Proof.
unfold add, w_max, max_elt in *.
unfolder; unfolder in MAX; desc; splits; try done.
by apply ACTS in MAX; destruct MAX; type_solver.
ins.
eapply MAX0.
apply CO1; [done|].
unfold add_co; basic_solver 12.
Qed.

Lemma w_max_add_write' G G' t n x l (WF: Wf G) 
  (WRITE: is_w_l l) w (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  w_max G' x  ≡₁ eq (ThreadEvent t n l).
Proof.
unfolder; ins; splits; ins.
- eapply max_w_unique with (G:=G'); try edone.
eapply add_preserves_wf; edone.
eapply w_max_add_write; try edone.
- subst. eby eapply w_max_add_write.
Qed.

Lemma add_max_non_write_fr G G' t n l w (WF: Wf G) (NWRITE: ~ is_w_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  G'.(fr) ≡ G.(fr).
Proof.
split; [|eapply add_fr_monotone; edone].
unfold fr.
rewrite add_non_write_co; try edone.
rewrite add_rf_in; try edone.
unfolder; ins; desf; splits; eauto.
exfalso; unfold w_max in *; unfolder in *; desf; eauto.
Qed.

Lemma add_max_write_co G G' t n l w (WF: Wf G) (WRITE: is_w_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  G'.(co) ≡ G.(co) ∪ ((G.(acts_set) ∩₁ is_w ∩₁ is_loc x) × (eq (ThreadEvent t n l))).
Proof.
rewrite add_write_co; eauto.
unfold add_co, w_max in *.
unfolder in MAX; desc.
split; unfolder; unfold is_loc; ins; desf; eauto.
- right; splits; eauto.
apply (wf_coE WF) in H; unfolder in H; desf.
apply (wf_coD WF) in H; unfolder in H; desf.
apply (wf_col WF) in H; unfolder in H; desf.
- exfalso; eauto.
- destruct (classic (x0=w)); subst; eauto 10.
  eapply (wf_co_total WF) in H0; try basic_solver.
desf; eauto 10.
exfalso; eauto.
Qed.

Lemma add_max_non_read_co_hb G G' t n l w (WF: Wf G) (NREAD: ~ is_r_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  co G' ⨾ (hb G')^? ≡ co G ⨾ (hb G)^?
   ∪ (dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid t⦘) 
     ∪₁ (acts_set G ∩₁ W ∩₁ is_loc x)) × eq (ThreadEvent t n l).
Proof.
rewrite add_max_write_co; try edone.
2: type_solver.
rewrite add_non_read_hb_alt_init; try edone.
split.
- rewrite (wf_co_init WF); try edone.
unfolder; ins; desf; eauto 21.
exfalso; eapply add_fresh; try edone.
apply (wf_hbE WF) in H0; unfolder in H0; desf.
- rewrite (dom_r (wf_coE WF)) at 2.
rewrite (dom_r (wf_hbE WF)) at 2.
basic_solver 21.
Qed.

Lemma add_max_non_read_dom_co_hb G G' t n l w (WF: Wf G) 
  (NREAD: ~ is_r_l l) 
  x (MAX: w_max G x w) (ADD: add G G' t n l w) 
  p :
  dom_rel (co G' ⨾ (hb G')^? ⨾ ⦗is_tid p⦘) ≡₁
  dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid p⦘) ∪₁ 
  (acts_set G ∩₁ W ∩₁ is_loc x ∩₁ (fun _ => t = p)).
Proof.
rewrite <- seqA.
rewrite add_max_non_read_co_hb; try edone.
unfold is_tid in *; basic_solver 12.
Qed.

Lemma add_max_write_co_hb G G' t n l w (WF: Wf G) 
  (WRITE: is_w_l l) 
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  co G' ⨾ (hb G')^? ≡ co G ⨾ (hb G)^?
   ∪ (dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid t⦘) 
      ∪₁ (acts_set G ∩₁ W ∩₁ is_loc x)
      ∪₁ (dom_rel (co G ⨾ (hb G)^? ⨾ ⦗eq w⦘) ∩₁ (fun _ => is_r_l l))) 
      × eq (ThreadEvent t n l).
Proof.
rewrite add_max_write_co; try edone.
destruct (is_r_l l) eqn:RR.
- rewrite add_read_hb_alt_init; try edone.
split.
+ rewrite (wf_co_init WF); try edone.
unfolder; ins; desf; eauto 21.
exfalso; eapply add_fresh; try edone.
apply (wf_hbE WF) in H0; unfolder in H0; desf.
+ rewrite (dom_r (wf_coE WF)) at 2.
rewrite (dom_r (wf_hbE WF)) at 2.
basic_solver 21.
- rewrite add_non_read_hb_alt_init; try edone.
2: type_solver.
split.
+ rewrite (wf_co_init WF); try edone.
unfolder; ins; desf; eauto 21.
exfalso; eapply add_fresh; try edone.
apply (wf_hbE WF) in H0; unfolder in H0; desf.
+ rewrite (dom_r (wf_coE WF)) at 2.
rewrite (dom_r (wf_hbE WF)) at 2.
basic_solver 21.
Qed.

Lemma add_max_write_dom_co_hb G G' t n l w (WF: Wf G) 
  (WRITE: is_w_l l) 
  x (MAX: w_max G x w) (ADD: add G G' t n l w) 
  p :
  dom_rel (co G' ⨾ (hb G')^? ⨾ ⦗is_tid p⦘) ≡₁
  dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid p⦘) ∪₁ 
  dom_rel (co G ⨾ (hb G)^? ⨾ ⦗eq w⦘) ∩₁ (fun _ => t = p /\ is_r_l l) ∪₁ 
  acts_set G ∩₁ W ∩₁ is_loc x ∩₁ (fun _ => t = p).
Proof.
rewrite <- seqA.
rewrite add_max_write_co_hb; try edone.
unfold is_tid; basic_solver 21.
Qed.

Lemma add_max_non_read_fr G G' t n l w (WF: Wf G) (COMP: complete G)
  (NREAD: ~ is_r_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  G'.(fr) ≡ G.(fr) ∪ ((G.(acts_set) ∩₁ is_r ∩₁ is_loc x) × (eq (ThreadEvent t n l))).
Proof.
unfold fr.
rewrite add_max_write_co; eauto.
rewrite add_non_read_rf; eauto.
unfolder; unfold is_loc; split; ins; desf; eauto.

- right; splits; eauto.
apply (wf_rfE WF) in H; unfolder in H; desf.
apply (wf_rfD WF) in H; unfolder in H; desf.
apply (wf_rfl WF) in H; unfolder in H; desf.
-
assert (A: exists z, rf G z x0) by (apply COMP; basic_solver).
desf. splits; [|by intro; desf].
eexists; splits; eauto.
right; splits; eauto.
apply (wf_rfE WF) in A; unfolder in A; desf.
apply (wf_rfD WF) in A; unfolder in A; desf.
apply (wf_rfl WF) in A; unfolder in A; desf.
- type_solver.
Qed.

Lemma add_max_rmw_fr G G' t n l w (WF: Wf G) (COMP: complete G)
  (WRITE: is_w_l l) (READ: is_r_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  G'.(fr) ≡ G.(fr) ∪ ((G.(acts_set) ∩₁ is_r ∩₁ is_loc x) × (eq (ThreadEvent t n l))).
Proof.
unfold fr.
rewrite add_max_write_co; eauto.
rewrite add_read_rf; eauto.
unfold w_max in *.
unfolder in MAX; desc.
unfolder; split; ins; desf; eauto 10.
- exfalso; eauto.
- right; splits; eauto.
apply (wf_rfE WF) in H; unfolder in H; desf.
apply (wf_rfD WF) in H; unfolder in H; desf.
apply (wf_rfl WF) in H; unfolder in H; desf.
unfold same_loc in *; congruence.
- exfalso; eauto.
-
assert (A: exists z, rf G z x0) by (apply COMP; basic_solver).
desf. splits.
+ eexists; splits; eauto.
right; splits; eauto.
apply (wf_rfE WF) in A; unfolder in A; desf.
apply (wf_rfD WF) in A; unfolder in A; desf.
apply (wf_rfl WF) in A; unfolder in A; desf.
unfold same_loc in *; congruence.
+ intro; desf.
eapply add_fresh; [edone|].
eapply domb_helper; try edone; rewrite (wf_rfE WF); basic_solver.
Qed.

Lemma add_max_non_write_hb_sc G G' t n l w (WF: Wf G) 
  (NWRITE: ~ is_w_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  (G'.(hb_sc)) ≡ G.(hb_sc) ∪
  (dom_rel ((hb_sc G)^? ⨾ ⦗ is_init ∪₁ G.(acts_set) ∩₁ is_tid t ∪₁ eq w ⦘) × (eq (ThreadEvent t n l))).
Proof.
unfold hb_sc.
rewrite add_sb, add_read_rf, add_non_write_co, add_max_non_write_fr; try edone.
arewrite (sb G
  ∪ (is_init ∪₁ acts_set G ∩₁ is_tid t)
    × eq (ThreadEvent t n l) ∪ (rf G ∪ singl_rel w (ThreadEvent t n l))
 ∪ co G ∪ fr G ≡
(sb G ∪ rf G ∪ co G ∪ fr G)
  ∪ ((is_init ∪₁ acts_set G ∩₁ is_tid t)
    × eq (ThreadEvent t n l) ∪ singl_rel w (ThreadEvent t n l))).
basic_solver 12.
rels.
rewrite ct_union_max; [| |basic_solver].
- clear; basic_solver 12.
- rewrite (dom_l (@wf_sbE G)), (dom_l (wf_rfE WF)), (dom_l (wf_coE WF)), (dom_l (wf_frE WF)).
apply dom_max_elt; unfolder; intro; desf; eapply add_fresh; eauto.
- type_solver.
Qed.

Lemma add_max_non_read_hb_sc G G' t n l w (WF: Wf G) (COMP: complete G)
  (NREAD: ~ is_r_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  (G'.(hb_sc)) ≡ G.(hb_sc) ∪
  (dom_rel ((hb_sc G)^? ⨾ 
  ⦗ is_init ∪₁ G.(acts_set) ∩₁ is_tid t ∪₁ G.(acts_set) ∩₁ is_loc x⦘) × (eq (ThreadEvent t n l))).
Proof.
unfold hb_sc.
rewrite add_sb, add_non_read_rf, add_max_write_co, add_max_non_read_fr; try edone.
arewrite (sb G ∪ (is_init ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l) ∪ rf G
 ∪ (co G
    ∪ (acts_set G ∩₁ is_w ∩₁ is_loc x)
      × eq (ThreadEvent t n l))
 ∪ (fr G
    ∪ (acts_set G ∩₁ is_r ∩₁ is_loc x)
      × eq (ThreadEvent t n l)) ≡
(sb G ∪ rf G ∪ co G ∪ fr G) ∪
((is_init ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l) ∪
(acts_set G ∩₁ is_w ∩₁ is_loc x)
      × eq (ThreadEvent t n l)
    ∪ (acts_set G ∩₁ is_r ∩₁ is_loc x)
      × eq (ThreadEvent t n l))).
basic_solver 12.
rels.
rewrite ct_union_max; [| |basic_solver].
- arewrite (acts_set G ∩₁ is_loc x ≡₁ acts_set G ∩₁ (is_r ∪₁ is_w) ∩₁ is_loc x).
by type_solver.
basic_solver 12.
- rewrite (dom_l (@wf_sbE G)), (dom_l (wf_rfE WF)), (dom_l (wf_coE WF)), (dom_l (wf_frE WF)).
apply dom_max_elt; unfolder; intro; desf; eapply add_fresh; eauto.
- type_solver.
Qed.

Lemma add_max_rmw_hb_sc G G' t n l w (WF: Wf G) (COMP: complete G)
  (WRITE: is_w_l l) (READ: is_r_l l)
  x (MAX: w_max G x w) (ADD: add G G' t n l w) : 
  (G'.(hb_sc)) ≡ G.(hb_sc) ∪
  (dom_rel ((hb_sc G)^? ⨾ 
  ⦗ is_init ∪₁ G.(acts_set) ∩₁ is_tid t ∪₁ G.(acts_set) ∩₁ is_loc x ∪₁ eq w⦘) × (eq (ThreadEvent t n l))).
Proof.
unfold hb_sc.
rewrite add_sb, add_read_rf, add_max_write_co, add_max_rmw_fr; try edone.
arewrite (sb G ∪ (is_init ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l) ∪ 
 (rf G ∪ singl_rel w (ThreadEvent t n l))
 ∪ (co G
    ∪ (acts_set G ∩₁ is_w ∩₁ is_loc x)
      × eq (ThreadEvent t n l))
 ∪ (fr G
    ∪ (acts_set G ∩₁ is_r ∩₁ is_loc x)
      × eq (ThreadEvent t n l)) ≡
(sb G ∪ rf G ∪ co G ∪ fr G) ∪
((is_init ∪₁ acts_set G ∩₁ is_tid t) × eq (ThreadEvent t n l) ∪
singl_rel w (ThreadEvent t n l) ∪
(acts_set G ∩₁ is_w ∩₁ is_loc x)
      × eq (ThreadEvent t n l)
    ∪ (acts_set G ∩₁ is_r ∩₁ is_loc x)
      × eq (ThreadEvent t n l))).
basic_solver 12.
rels.
rewrite ct_union_max; [| |basic_solver].
- arewrite (acts_set G ∩₁ is_loc x ≡₁ acts_set G ∩₁ (is_r ∪₁ is_w) ∩₁ is_loc x).
by type_solver.
basic_solver 12.
- rewrite (dom_l (@wf_sbE G)), (dom_l (wf_rfE WF)), (dom_l (wf_coE WF)), (dom_l (wf_frE WF)).
apply dom_max_elt; unfolder; intro; desf; eapply add_fresh; eauto.
Qed.

End ADD_MAX.
