(******************************************************************************)
(** * SRAG simulates SRAL: rmw step   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import AuxRel.
Require Import Basic.
Require Import Filters.
Require Import Labels Events Execution.
Require Import SRAL Wmax SRAG.
Require Import Write_lists Match.

Require Import FG.

Require Import GsimL_read GsimL_write.
Require Import LsimG_read.

Set Implicit Arguments.
Remove Hints plus_n_O.


Lemma GsimL_rmw_pick_w st st'' (WF_S: wf_state st'')
          (G : execution)
      t x vr vw (SRAL: SRAL_step st st'' (Some (t, (Armw x vr vw))))
          (MATCH: SGmatch st G) : 
          exists e, w_max G x e /\ valw e = vr.
Proof.
red in MATCH; desc.
destruct (WF_S t) as [ol OL_IN].
inversion SRAL; desf.
red in READ. red in WRITE. desc.
apply P in OL_IN. apply T0 in OL_IN. apply SMATCH in OL_IN; desc.
exists (wl 0).
clear - OL_IN.
red in OL_IN; desc. red in WL_WF; desc.
splits.
- red; unfolder; splits; eauto; ins.
  eapply WL_G; red; exists 0; splits; eauto; ins; omega.
  eapply WL_W; red; exists 0; splits; eauto; ins; omega.
  rewrite WL_LOC; ins; omega.
  eapply WL_RMW with (n:=0); ins;[omega|basic_solver].
- rewrite WL_VAL; ins; omega.
Qed.

Lemma GsimL_rmw (st st'' : State) (G : execution)
      t x vr vw (SRAL: SRAL_step st st'' (Some (t, (Armw x vr vw))))
      (MATCH: SGmatch st G) : 
         exists G', SRAG_step G G' (Some (t, (Armw x vr vw))) /\ 
                    SGmatch st'' G'.
Proof.
forward (eapply GsimL_rmw_pick_w); try edone.
eapply SRAL_step_preserves_wf; [eapply MATCH|]; eauto.
intro MAX; desc.
cdes MAX; unfolder in MAX1; desc.
inversion SRAL; subst; clear SRAL.

remember (valw e) as vr.
remember (loc e) as x.

forward (eapply exists_add' with (w:=e) (l:= Armw x vr vw) (t:=t)); try edone.
apply MATCH.
intro RMW_ADD; destruct RMW_ADD as [n [G' RMW_ADD]].
exists G'; splits.
by red; eexists _,_,_,_; splits; eauto 21; apply MATCH.

red; splits; try done.
eapply add_preserves_wf; try edone.
apply MATCH.

intros p ol'' IN''.
red in WRITE; desc.
specialize (P _ _ IN'').
specialize (T _ _ IN'').
red in READ; desc.

red in MATCH; desc.
assert (SMATCH' := SMATCH).
specialize (SMATCH' _ _ (T0 _ T)); desc.

assert (wl 0 = e).
{ eapply max_w_unique; try edone.
  red in SMATCH'; desc.
  red in WL_WF; desc.
  red; unfolder; splits; eauto.
  eapply WL_G; red; exists 0; splits; eauto; ins; omega.
  eapply WL_W; red; exists 0; splits; eauto; ins; omega.
  rewrite WL_LOC; ins; omega.
  ins.
  eapply WL_RMW with (n:=0).
  omega.
  ins.
  basic_solver. }

assert (SGmatch_wl G t (filter_ind_pre ol'' (I p ol'')) (fun n => wl (S n))).
{ forward (eapply exists_add' with (w:=e) (l:= Aload x vr) (t:=t)); try edone.
  intro RADD; destruct RADD as [nr [Gr RADD]].
  eapply LsimG_read_wl; try edone.
  subst.
  eapply GsimL_read_helper_t with (G':=Gr) (tw:=tw); try edone.
  - red in SMATCH'; desc.
    red in WL_WF; desc.
    rewrite WL_TID; ins.
    omega.
  - by ins; apply dom_max_elt. }

destruct (classic (p <> t)) as [N|E].
- specialize (SMATCH _ _ (P0 p N _ P)); desc.
  eexists; eapply GsimL_write_helper1; eauto.
  ins; desf.
  red in SMATCH'; desc.
  red in WL_CONS; desc.
  intro; eapply WL_O.
  unfolder; eexists; splits; eauto.
  unfold order_wl; exists 0, (S n0); splits; eauto; simpl; omega.
- assert (p = t) by tauto; subst p.
  specialize (SMATCH _ _ (T0 _ P)); desc.
  assert (wl0 0 = e).
  { eapply max_w_unique; try edone.
    red in SMATCH; desc.
    red in WL_WF; desc.
    red; unfolder; splits; eauto.
    eapply WL_G; red; exists 0; splits; eauto; ins; omega.
    eapply WL_W; red; exists 0; splits; eauto; ins; omega.
    rewrite WL_LOC; ins; omega.
    ins.
    eapply WL_RMW with (n:=0).
    omega.
    ins.
    basic_solver. }

  eexists; eapply GsimL_write_helper1; eauto.
  + forward (eapply exists_add' with (w:=e) (l:= Aload x vr) (t:=t)); try edone.
    intro RADD; destruct RADD as [nr [Gr RADD]].
    eapply LsimG_read_wl; try edone.
    subst.
    eapply GsimL_read_helper_t with (tw:=tw); try edone.
    * red in SMATCH'; desc.
     red in WL_WF; desc.
     rewrite WL_TID; ins.
     omega.
    * by ins; apply dom_max_elt.
    * by rewrite H1.
  + ins; desf.
    red in SMATCH; desc.
    red in WL_CONS; desc.
    intro; eapply WL_O.
    unfolder; eexists; splits; eauto.
    unfold order_wl; exists 0, (S n0); splits; eauto; simpl; omega.
  + ins; desf.
    red in SMATCH'; desc.
    red in WL_CONS; desc.
    intro; eapply WL_O.
    unfolder; eexists; splits; eauto.
    unfold order_wl; exists 0, (S n0); splits; eauto; simpl; omega.
Qed.


