(******************************************************************************)
(** * SRAG simulates SRAL: write step   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import AuxRel.
Require Import Basic.
Require Import Filters.
Require Import Labels Events Execution.
Require Import SRAL Wmax Add_max SRAG.
Require Import Write_lists Match.

Require Import FG.

Set Implicit Arguments.
Remove Hints plus_n_O.

Definition make_wl (wl_p wl_t : nat -> Event) (P : list bool) 
                 (w : Event) (fp ft : nat -> nat) 
                 (rel : relation Event) n : Event :=
 if (Nat.leb (length P) n) then default_event
 else if (Nat.ltb n (min_of_P P)) then (wl_p (fp n))
      else if (nth n P false) then w
           else if (excluded_middle_informative 
  (rel (wl_p (fp n)) (wl_t (ft n)))) then (wl_t (ft n))
           else (wl_p (fp n)).

Lemma make_wl_cases wl_p wl_t P w fp ft rel n
      a (EQ: make_wl wl_p wl_t P w fp ft rel n = a)
      (LT: n < length P) :
      a = w /\ nth n P false = true \/
      (a = wl_p (fp n) /\ nth n P false = false /\
        (n < min_of_P P \/
         min_of_P P <= n /\
         ~ rel (wl_p (fp n)) (wl_t (ft n)))) \/
      a = wl_t (ft n) /\ (min_of_P P <= n) /\ 
        nth n P false = false /\ rel (wl_p (fp n)) (wl_t (ft n)).
Proof.
generalize (@false_before_min P n); ins.
unfold make_wl in *; desf.
- apply Nat.leb_le in Heq; omega.
- apply Nat.ltb_lt in Heq0; auto 7.
- apply Nat.ltb_ge in Heq0; auto.
- apply Nat.ltb_ge in Heq0; auto 8.
- apply Nat.ltb_ge in Heq0; auto 8.
Qed.

Lemma make_wl_cases_weak wl_p wl_t P w fp ft rel n
      a (EQ: make_wl wl_p wl_t P w fp ft rel n = a)
      (LT: n < length P) :
      a = w /\ nth n P false = true \/
      (a = wl_p (fp n) /\ nth n P false = false) \/
      a = wl_t (ft n) /\ (min_of_P P <= n) /\ 
        nth n P false = false.
Proof.
apply make_wl_cases in EQ; desf; eauto.
Qed.

Lemma in_wl_make_wl wl_p wl_t P w rel (ol' : list Opt) 
  ol_p ol_t 
  (LEN: length ol' = length P)
  (LP: ol_p = filter_ind  ol' P) 
  (LT: ol_t = filter_ind_pre  ol' P) 
  fp (FP : fp = filter_ind_map P)
  ft (FT : ft = filter_ind_pre_map P) :
  in_wl (make_wl wl_p wl_t P w fp ft rel) (length ol') ⊆₁ 
  in_wl wl_p (length ol_p) ∪₁ in_wl wl_t (length ol_t) ∪₁ (eq w).
Proof.

assert (LEN_FP: forall n, n < length ol' -> 
  n < min_of_P P \/ nth n P false = false -> fp n < length ol_p).
 by ins; desf; apply filter_ind_map_length; auto;
    desf; apply false_before_min.

assert (LEN_FT: forall n, n < length ol' -> n >= min_of_P P ->
 nth n P false = false -> ft n < length ol_t).
by ins; desf; apply filter_ind_pre_map_length; auto.

assert (FP_IN: forall n0 
  (INa: n0 < length ol')
  (H: n0 < min_of_P P \/ nth n0 P false = false),
  in_wl wl_p (length ol_p) (wl_p (fp n0))).
by ins; red; eauto.

assert (FT_IN: forall n0 
  (INa: n0 < length ol')
  (Heq0: n0 >= min_of_P P)
  (Heq1: nth n0 P false = false),
  in_wl wl_t (length ol_t) (wl_t (ft n0))).
by ins; red; eauto.

unfolder; unfold in_wl; intros a INa; desc.
apply make_wl_cases_weak in INa0; desf; try omega; desf; eauto 12.
Qed.

Lemma GsimL_write_helper1 
  (G : execution) (WF_G: Wf G)
  t x v n w l e (E: e = (ThreadEvent t n l))
  G' (MAX: w_max G x w) (ADD: add G G' t n l w)
  P p ol'
  ol_p (OL_P: ol_p = filter_ind ol' P)
  ol_t (OL_T: ol_t = filter_ind_pre ol' P)
  (T_LOC : forall o (INo: In o ol_t), loc_opt o <> x)
  (T_LOCt : p = t -> forall o (INo: In o ol_p), loc_opt o <> x)
  (NU : forall o (INo: In o ol_p) (LOCo: loc_opt o = x),
       rmw_flag_opt o <> true)
  (I_WF : wf_ind_choice_ol ol' P t x v)
  fp (FP : fp = filter_ind_map P)
  ft (FT : ft = filter_ind_pre_map P)
  wl_p (MATCH_WL_P: SGmatch_wl G p ol_p wl_p)
  wl_t (MATCH_WL_T: SGmatch_wl G t ol_t wl_t)
  (Wl: is_w_l l) (Wloc: loc_l l = x) (Wval: valw_l l = v)
  (Rl_wl_p : p = t -> is_r_l l -> forall n, n < length ol_p -> 
    ~ (co G ⨾ (hb G)^?) (wl_p n) w)
  (Rl_wl_t : is_r_l l -> forall n, n < length ol_t ->  
     ~ (co G ⨾ (hb G)^?) (wl_t n) w) :
  SGmatch_wl G' p ol' (make_wl wl_p wl_t P e fp ft G.(co)). 
Proof.
destruct MATCH_WL_P as [WF_WL_P CONS_WL_P]; unnw.
destruct MATCH_WL_T as [WF_WL_T CONS_WL_T]; unnw.
destruct WF_WL_P as [WL_P_W [WL_P_G [WL_P_TID [WL_P_LOC [WL_P_VAL WL_P_RMW]]]]]; unnw. 
destruct WF_WL_T as [WL_T_W [WL_T_G [WL_T_TID [WL_T_LOC [WL_T_VAL WL_T_RMW]]]]]; unnw. 
red in I_WF; desc.


assert (LEN_FP: forall n, n < length ol' -> 
  n < min_of_P P \/ nth n P false = false -> fp n < length ol_p).
by ins; desf; apply filter_ind_map_length; auto;
    desf; apply false_before_min.

assert (LEN_FT: forall n, n < length ol' -> n >= min_of_P P ->
 nth n P false = false -> ft n < length ol_t).
by ins; desf; apply filter_ind_pre_map_length; auto.

assert (FP_IN: forall n0 
  (INa: n0 < length ol')
  (H: n0 < min_of_P P \/ nth n0 P false = false),
  in_wl wl_p (length ol_p) (wl_p (fp n0))).
by ins; red; eauto.

assert (FT_IN: forall n0 
  (INa: n0 < length ol')
  (Heq0: n0 >= min_of_P P)
  (Heq1: nth n0 P false = false),
  in_wl wl_t (length ol_t) (wl_t (ft n0))).
by ins; red; eauto.

cdes MAX.
red; splits; red; splits; try done.
- rewrite in_wl_make_wl, WL_P_W, WL_T_W; basic_solver.
- rewrite in_wl_make_wl, WL_P_G, WL_T_G; try edone.
  rewrite (add_acts_set ADD); basic_solver 12.
- ins.
  remember (make_wl wl_p wl_t P e fp ft (co G) n0) as a.
  symmetry in Heqa.
  apply make_wl_cases_weak in Heqa; try omega; desf.
  + rewrite (P_TID n0); ins.
  + erewrite WL_P_TID; eauto.
    by erewrite <- filter_ind_map_prop'; auto.
  + erewrite WL_T_TID; eauto.
    by erewrite <- filter_ind_pre_map_prop'; auto.
- ins.
  remember (make_wl wl_p wl_t P e fp ft (co G) n0) as a.
  symmetry in Heqa.
  apply make_wl_cases_weak in Heqa; try omega; desf.
  + rewrite (P_LOC n0); ins.
  + erewrite WL_P_LOC; eauto.
    by erewrite <- filter_ind_map_prop'; auto.
  + erewrite WL_T_LOC; eauto.
    by erewrite <- filter_ind_pre_map_prop'; auto.
- ins.
  remember (make_wl wl_p wl_t P e fp ft (co G) n0) as a.
  symmetry in Heqa.
  apply make_wl_cases_weak in Heqa; try omega; desf.
  + rewrite (P_VAL n0); ins.
  + erewrite WL_P_VAL; eauto.
    by erewrite <- filter_ind_map_prop'; auto.
  + erewrite WL_T_VAL; eauto.
    by erewrite <- filter_ind_pre_map_prop'; auto.
- ins.
  remember (make_wl wl_p wl_t P e fp ft (co G) n0) as a.
  unfolder; intro; desc.
  eapply add_max_write_co in H; try edone.
  destruct H as [H|H].
  + symmetry in Heqa.
    apply make_wl_cases_weak in Heqa; try omega; desf.
    * hahn_rewrite (dom_l (wf_coE WF_G)) in H.
      eapply (add_fresh ADD); unfolder in H; desf.
    * eapply WL_P_RMW; eauto.
      by erewrite <- filter_ind_map_prop'; auto.
      by basic_solver.
    * eapply WL_T_RMW; eauto.
      by erewrite <- filter_ind_pre_map_prop'; auto.
      by basic_solver.
  + unfolder in H; desc.
    symmetry in Heqa.
    apply make_wl_cases_weak in Heqa; try omega; desf.
    * eapply (add_fresh ADD); unfolder in H; desf.
    * eapply NU; eauto.
      erewrite filter_ind_map_prop'; eauto.
      eapply nth_In; eauto.
      unfold is_loc in H1.
      erewrite WL_P_LOC in H1; eauto.
      eby erewrite filter_ind_map_prop'; auto.
    * eapply NU; eauto.
      erewrite filter_ind_map_prop'; eauto.
      eapply nth_In; eauto.
      unfold is_loc in H1.
      erewrite WL_T_LOC in H1; eauto.
      eby erewrite filter_ind_pre_map_prop'; auto.
- rewrite add_max_write_dom_co_hb; try edone.
  unfold set_subset, set_inter.
  unfold in_wl.
  ins; exfalso; desc.
  destruct CONS_WL_P as [CONS_WL_P _]; unnw.
  assert (x0 <> e).
  { clear H1; intro; subst.
    eapply (add_fresh ADD).
    unfolder in H0; ins; desf;
    apply (dom_l (wf_coE WF_G)) in H0; 
    revert H0; basic_solver.
  } 
  unfold set_union in H0; destruct H0 as [[H0|H0]|X].
  + apply make_wl_cases in H1; try omega; desf.
    by eapply CONS_WL_P; split; [eauto|basic_solver 12].
    by eapply CONS_WL_P; split; [eauto|basic_solver 12].
    eapply CONS_WL_P; split; eauto.
    by eapply dom_rel_trans; try edone; apply WF_G.
  + desc; subst p.
    apply make_wl_cases in H1; try omega.
    destruct H1 as [H1|[H1|H1]]; desc.
    by desf.
    eapply Rl_wl_p; eauto; revert H0; basic_solver 12.
    eapply Rl_wl_p; eauto. subst x0.
    eapply dom_rel_trans in H0; try edone; try apply WF_G.
    revert H0; basic_solver 21.
  + desc; subst p.
    eapply make_wl_cases_weak in H1; try omega.
    destruct H1; desc; eauto.
    destruct H0; desc; subst x0.
    * eapply T_LOCt; eauto.
      eapply nth_In; eauto.
      rewrite <- WL_P_LOC; eauto.
    * eapply T_LOC; eauto.
      eapply nth_In; eauto.
      rewrite <- WL_T_LOC; eauto.
- rewrite add_max_write_co_hb; try edone.
  unfold order_wl.
  unfold seq at 1; red; ins. desc.
  apply make_wl_cases in H3; [|omega].
  destruct H3 as [XX|Z]; desc.
  by eapply (add_fresh ADD);
      unfolder in H0; ins; desf;
      apply (dom_l (wf_coE WF_G)) in H0; 
      revert H0; basic_solver.
  apply make_wl_cases in H2; [|omega].
  rename H2 into X.
  assert (H1': n1 < length ol') by omega.
  rename H0 into B.
  destruct B as [B|B].

  + destruct X as [XX|X].
    by hahn_rewrite (dom_r (wf_coE WF_G)) in B;
       hahn_rewrite (dom_r (wf_hbE WF_G)) in B; 
       eapply (add_fresh ADD);
       revert B; basic_solver.
    destruct CONS_WL_P as [_ CONS_WL_P]; unnw.
    destruct CONS_WL_T as [_ CONS_WL_T]; unnw.
    unfold order_wl in *.
    destruct X,Z; desc.
    * eapply CONS_WL_P.
      eexists; split; eauto.
      exists (fp n1), (fp n2); splits; eauto.
      by desf; apply filter_ind_map_monotone_strict; try done; omega.
    * assert (CO: (co G ⨾ (hb G)^?) (wl_p (fp n2)) (wl_p (fp n1))).
      by generalize (co_trans WF_G); revert H5 B; basic_solver 12.
      clear H5 B.
      eapply CONS_WL_P.
      eexists; split; eauto.
      exists (fp n1), (fp n2); splits; eauto.
      by desf; apply filter_ind_map_monotone_strict; try done; omega.
    * destruct H4. by omega. desc. subst x0 z.
      assert (CO: (co G ⨾ (hb G)^?) (wl_t (ft n2)) (wl_t (ft n1))).
      { cut ((co G)^? (wl_t (ft n2)) (wl_p (fp n2))).
        - generalize (co_trans WF_G); revert B.
          remember (wl_p (fp n2)) as a.
          remember (wl_t (ft n1)) as b.
          remember (wl_t (ft n2)) as c.
          clear; basic_solver 21.
        - eapply tot_ex'; eauto.
          eapply WF_G.
          + basic_solver.
          + unfolder; splits; eauto.
            erewrite WL_P_LOC, WL_T_LOC; eauto.
            desf; rewrite <- filter_ind_pre_map_prop'; eauto.
            rewrite <- filter_ind_map_prop'; try edone.
      } 
      eapply CONS_WL_T.
      eexists; split; eauto.
      exists (ft n1), (ft n2); splits; eauto.
      by desf; apply filter_ind_pre_map_monotone_strict; try done; omega.

    * eapply CONS_WL_T.
      eexists; split; eauto.
      exists (ft n1), (ft n2); splits; eauto.
      by desf; apply filter_ind_pre_map_monotone_strict; try done; omega.

  + unfold cross_rel in B; desc.
    destruct X as [X|XX]; cycle 1.
    by clear Z; eapply (add_fresh ADD); desf; rewrite XX; eauto.
    destruct Z as [Z|Z]; cycle 1; desc.

    * unfold set_union in B; destruct B as [[B|B]|B]; subst x0 z.
      -- destruct CONS_WL_T as [CONS_WL_T _]; unnw.
         eapply CONS_WL_T; revert B; basic_solver 12.
      -- eapply T_LOC. 
         by eapply nth_In; eauto.
         rewrite <- WL_T_LOC; eauto.
         revert B; basic_solver.
      -- eapply Rl_wl_t; eauto; revert B; basic_solver 12.
    * destruct Z1 as [Z1|Z1]; desc.
      by destruct (classic (n1 < min_of_P P));
        [apply false_before_min in H0; desf| omega].
      assert (CO: (co G)^? (wl_t (ft n2)) (wl_p (fp n2))).
      { eapply tot_ex'.
        eapply WF_G.
        + basic_solver.
        + unfolder; splits; eauto.
          erewrite WL_P_LOC, WL_T_LOC; eauto.
          desf; rewrite <- filter_ind_pre_map_prop'; try edone.
          rewrite <- filter_ind_map_prop'; try edone.
        + done.
      }
      unfold set_union in B; destruct B as [[B|B]|B]; subst x0 z.
      -- assert (CO': dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid t⦘) (wl_t (ft n2))).
         { remember (wl_t (ft n2)) as a.
           remember (wl_p (fp n2)) as b.
           generalize (co_trans WF_G); revert CO B.
           clear; basic_solver 21.
         }
         clear CO B.
         destruct CONS_WL_T as [CONS_WL_T _]; unnw.
         eapply CONS_WL_T; split; eauto.
      -- eapply T_LOC. 
         by eapply nth_In; eauto.
         rewrite <- WL_T_LOC; eauto.
         replace (loc (wl_t (ft n2))) with (loc (wl_p (fp n2))).
         revert B; basic_solver 12.
         by symmetry; unfolder in CO; desf; [congruence| apply (wf_col WF_G)].
     -- assert (CO': (dom_rel (co G ⨾ (hb G)^? ⨾ ⦗eq w⦘)
              ∩₁ (fun _ : Event => is_r_l l)) (wl_t (ft n2))).
         { remember (wl_t (ft n2)) as a.
           remember (wl_p (fp n2)) as b.
           generalize (co_trans WF_G); revert CO B.
           clear; basic_solver 21.
         }
         clear CO B.
         eapply Rl_wl_t; eauto; revert CO'; basic_solver 12.
Qed.

Lemma GsimL_write_helper (st st' : State) (WF_S: wf_state st')
          (G : execution) 
          t x v I (SRAL: SRAL_write_step_helper st st' t x v I)
          (MATCH: SGmatch st G)  
          n w G' l
          (MAX: w_max G x w) (ADD: add G G' t n l w)
          (Wl: is_w_l l) (Wloc: loc_l l = x) (Wval: valw_l l = v)
          (Rl: is_r_l l -> false ) :
             SGmatch st' G'.
Proof.
red in SRAL; ins; desc.
red; splits; try done.
eapply add_preserves_wf; try edone; by apply MATCH.
intros p ol' IN.
red in MATCH; desc.
assert (SMATCH' := SMATCH).
specialize (SMATCH _ _ (P _ _ IN)); desc.
specialize (SMATCH' _ _ (T _ _ IN)); desc.
eexists.
eapply GsimL_write_helper1; try eauto.
ins; desf; eauto. 
Qed.

Lemma GsimL_write (st st' : State) (G : execution)
      t x v (SRAL: SRAL_step st st' (Some (t, (Astore x v))))
      (MATCH: SGmatch st G) : 
         exists G', SRAG_step G G' (Some (t, (Astore x v))) /\ 
                    SGmatch st' G'.
Proof.
inversion SRAL; subst; clear SRAL.
assert (WF_G: Wf G) by apply MATCH.
forward (eapply (max_w_exists WF_G x)); intro MAX; desc.
forward (eapply exists_add' with (w:=w) (l:= Astore x v) (t:=t)); try edone.
all: try by red in MAX; revert MAX; basic_solver.
ins; desf.
exists G'; splits.
red; eauto 21.
eby eapply GsimL_write_helper.
Qed.
