(******************************************************************************)
(** * Definition of labels *)
(******************************************************************************)

Require Import FinFun ListDec.
Require Import List.
Require Import Hahn.

Set Implicit Arguments.
Remove Hints plus_n_O.

Parameter Loc : Type.

Parameter finite_loc : Finite Loc.

Parameter Loc_list : list Loc.
Parameter Loc_list_prop : NoDup Loc_list /\ Full Loc_list.
Parameter X0 : Loc.

Section Labels.

Definition Tid := nat.
Definition Val := nat.

(******************************************************************************)
(** ** Labels  *)
(******************************************************************************)

Inductive Lab := 
  | Aload  (x:Loc) (v:Val)
  | Astore (x:Loc) (v:Val)
  | Armw   (x:Loc) (vr vw:Val).

Definition loc_l l :=
  match l with
  | Aload  x _
  | Astore x _
  | Armw   x _ _ => x
  end.

Definition valr_l l :=
  match l with
  | Aload  _ v
  | Armw   _ v _ => v
  | _ => 0 
  end.

Definition valw_l l :=
  match l with
  | Astore _ v
  | Armw   _ _ v => v
  | _ => 0 
  end.

Definition is_r_l l := 
  match l with
  | Aload  _ _ => true
  | Armw   _ _ _ => true
  | _ => false
  end.

Definition is_w_l l := 
  match l with
  | Astore  _ _  => true
   | Armw   _ _ _ => true
  | _ => false
  end.

Definition is_rmw_l l := 
  match l with
  | Armw   _ _ _ => true
  | _ => false
  end.

Definition default_lab := Astore X0 0.

End Labels.

