(******************************************************************************)
(** * SRAL <-> SRAG simulation definitions   *)
(******************************************************************************)

Require Import Coq.Logic.ClassicalChoice.

Require Import Omega.
Require Import Hahn.

Require Import Basic.
Require Import Labels Events Execution.
Require Import SRAL SRAG.

Require Import FG.
Require Import Write_lists.

Set Implicit Arguments.
Remove Hints plus_n_O.

Section Match.

Definition make_r_opt e u := R_opt (tid e) (loc e) (valw e) u.

Definition wf_wl (wl : nat -> Event) N (ol : list Opt) (G : execution) :=
  ⟪ WL_W  : in_wl wl N ⊆₁ is_w ⟫ /\
  ⟪ WL_G  : in_wl wl N ⊆₁ G.(acts_set) ⟫ /\
  ⟪ WL_TID  : forall n (LEN: n < length ol), 
              tid (wl n) = tid_opt (nth n ol default_opt) ⟫ /\
  ⟪ WL_LOC  : forall n (LEN: n < length ol), 
              loc (wl n) = loc_opt (nth n ol default_opt) ⟫ /\
  ⟪ WL_VAL  : forall n (LEN: n < length ol), 
              valw (wl n) = val_opt (nth n ol default_opt) ⟫ /\
  ⟪ WL_RMW  : forall n (LEN: n < length ol)
               (RMWF: rmw_flag_opt (nth n ol default_opt) = true), 
                ~ dom_rel G.(co) (wl n) ⟫.

Definition consistent_wl (wl : nat -> Event) N G p :=
  ⟪ WL_D  : (in_wl wl N) ∩₁ dom_rel ( G.(co) ⨾ G.(hb)^? ⨾ ⦗ is_tid p ⦘) ⊆₁ ∅ ⟫ /\
  ⟪ WL_O  : irreflexive (order_wl wl N ⨾ G.(co) ⨾ G.(hb)^?) ⟫.

Definition SGmatch_wl G p ol wl :=
  ⟪ WL_WF : wf_wl wl (length ol) ol G ⟫ /\ 
  ⟪ WL_CONS : consistent_wl wl (length ol) G p ⟫.

Definition SGmatch (st : State) (G : execution) :=
  ⟪ WF_ST : wf_state st ⟫ /\
  ⟪ WF_G : Wf G ⟫ /\
  ⟪ SMATCH: forall p ol (IN: st p ol), 
           exists wl, SGmatch_wl G p ol wl ⟫.

Lemma SGmatch_SGmatch_fun st G (A: SGmatch st G): 
 exists fwl,  forall p ol (IN: st p ol), 
    wf_wl (fwl p ol) (length ol) ol G /\ 
    consistent_wl (fwl p ol) (length ol) G p.
Proof.
  intros; red in A; desc.
  apply (choice (fun p fwlp => forall (ol : list Opt),
    st p ol ->
    wf_wl (fwlp ol) (length ol) ol G /\ consistent_wl (fwlp ol) (length ol) G p)).
  intro p.
  apply (choice (fun ol yol => 
    st p ol ->
    wf_wl yol (length ol) ol G /\ consistent_wl yol (length ol) G p)); intro ol.
  destruct (classic (st p ol)).
  - edestruct SMATCH; eauto.
  - exists (fun _ => default_event); contradiction.
Qed.

(******************************************************************************)
(** * lemmas for initialization   *)
(******************************************************************************)

Lemma SGmatch_empty G (WF_G: Wf G): 
  SGmatch (fun p ol => ol = nil) G.
Proof.
red; splits; ins; subst.
by red; ins; eexists; eauto.
exists (fun n => default_event).
red; ins; splits; red; splits; ins.
all: try omega.
all: try by red; unfolder; unfold in_wl, order_wl; ins; desc; omega.
Qed.

Lemma SGmatch_init st (INIT: SRALinit st): 
  SGmatch st Ginit.
Proof.
red; ins; splits.
by apply INIT.
by apply Ginit_wf.
ins.
exists (fun n => InitEvent (loc_opt (nth n ol default_opt))).
red in INIT; desc.
specialize (INIT0 _ _ IN).
red; ins; splits; red; splits; ins.
- unfolder; unfold in_wl; ins; desc; subst x; type_solver.
- unfolder; unfold in_wl; ins; desc; subst x.
  apply Ginit_acts_set; ins.
- by specialize (INIT0 _ LEN); desc.
- by specialize (INIT0 _ LEN); desc.
- unfolder; intro; desf.
- unfolder; ins; desf.
- unfolder; ins; desf.
Qed.

Lemma SGmatch_G_init st (WF_S: wf_state st) (MATCH: SGmatch st Ginit) : 
  SRALinit st.
Proof.
split; ins.
red in MATCH; desc.
specialize (SMATCH _ _ IN); desc.
red in SMATCH; desc.
red in WL_WF; desc.
specialize (WL_TID _ LEN).
specialize (WL_VAL _ LEN).
rewrite <- WL_TID.
rewrite <- WL_VAL.
specialize (WL_G (wl n)).
forward (apply WL_G).
by red; eauto.
intro A.
apply Ginit_acts_set in A.
destruct (wl n); vauto.
Qed.

End Match.

