(******************************************************************************)
(** * SRAG memory subsystem   *)
(******************************************************************************)

Require Import Omega.

Require Import Hahn.
Require Import AuxRel.

Require Import Labels Events Execution FG Wmax.

Set Implicit Arguments.
Remove Hints plus_n_O.

Notation "'R'" := is_r.
Notation "'W'" := is_w.
Notation "'RMW'" := is_rmw.

Section SRAG.

Definition SRAGinit G := G = Ginit.

Definition SRAG_step G G' o :=
  exists t l n w,
  ⟪ O: o = Some (t,l) ⟫ /\
  ⟪ WF_G: Wf G ⟫ /\
  ⟪ MAX: is_w_l l -> w_max G (loc_l l) w ⟫ /\
  ⟪ ADD: add G G' t n l w ⟫ /\
  ⟪ CO: ~ is_w_l l -> ~ dom_rel (G.(co) ⨾ G.(hb)^? ⨾ ⦗ is_tid t ⦘) w ⟫.

(******************************************************************************)
(** * SRAG preserves properties preserved by adding events    *)
(******************************************************************************)

Lemma SRAG_step_preserves_wf G G' o
  (STEP: SRAG_step G G' o): Wf G'.
Proof.
red in STEP; desc.
eapply add_preserves_wf; edone.
Qed.

Lemma SRAG_step_preserves_complete G G' o (COMP: complete G) 
  (STEP: SRAG_step G G' o): complete G'.
Proof.
red in STEP; desc.
eapply add_preserves_complete; edone.
Qed.

Lemma SRAG_step_preserves_no_co_to_init G G' o
  (CO: G.(co) ≡ G.(co) ⨾  ⦗fun x => ~ is_init x⦘) 
  (STEP: SRAG_step G G' o): G'.(co) ≡ G'.(co) ⨾  ⦗fun x => ~ is_init x⦘.
Proof.
red in STEP; desc.
eapply add_preserves_no_co_to_init; edone.
Qed.

Lemma SRAG_step_preserves_rf_to_rmw_in_co G G' o
  (RF: G.(rf) ⨾ ⦗is_rmw⦘ ⊆ G.(co)) 
  (STEP: SRAG_step G G' o): G'.(rf) ⨾ ⦗is_rmw⦘ ⊆ G'.(co).
Proof.
red in STEP; desc.
eapply add_preserves_rf_to_rmw_in_co; edone.
Qed.

End SRAG.
