(******************************************************************************)
(** * Main result: SRAL and SRAG have the same runs     *)
(******************************************************************************)

Require Import Hahn.

Require Import Labels Events Execution FG.
Require Import SRAL SRAG.
Require Import Match.

Require Import GsimL_read GsimL_write GsimL_rmw GsimL_lose.
Require Import LsimG_read LsimG_write LsimG_rmw.

Require Import Memory_system.

Set Implicit Arguments.
Remove Hints plus_n_O.


Definition SRAL_run := MS_run SRALinit SRAL_step.
Definition SRAG_run := MS_run SRAGinit SRAG_step.

Lemma SRAL_in_SRAG : forall L, SRAL_run L -> SRAG_run L.
Proof.
eapply MS1_runs_in_MS2_runs_fwd with (sim_rel := SGmatch).
- ins; exists Ginit; splits.
  by vauto.
  by apply SGmatch_init.
- ins; destruct o.
  + left; destruct p, l.
    * eapply GsimL_read; try edone.
      by inversion H; subst; clear H.
      apply H0.
    * eapply GsimL_write; try edone.
    * eapply GsimL_rmw; try edone.
  + right; splits; auto.
    eapply GsimL_lose; try edone.
Qed.

Lemma SRAG_in_SRAL : forall L, SRAG_run L -> SRAL_run L.
Proof.
eapply MS1_runs_in_MS2_runs_bwd with 
  (sim_rel := fun x y => SGmatch y x).
- ins. eexists.
  apply SGmatch_empty.
  eapply MS_reachable_preserves; try edone.
  + ins; unfold SRAGinit in *; subst; apply Ginit_wf.
  + eby ins; eapply SRAG_step_preserves_wf.
- ins; unfold SRAGinit in *; subst. 
  apply SGmatch_G_init; try done.
  apply REL.
- ins; destruct o.
  + left; destruct p, l.
    * eapply LsimG_read; try edone.
      apply H0.
      eapply MS_reachable_preserves; try edone.
      -- ins; unfold SRAGinit in *; subst; apply Ginit_wf.
      -- eby ins; eapply SRAG_step_preserves_wf.
    * eapply LsimG_write; try edone.
    * eapply LsimG_rmw; try edone.
  + inversion H; desf.
Qed.

Lemma SRAL_eq_SRAG : forall L, SRAL_run L <-> SRAG_run L.
Proof.
ins; split; auto using SRAL_in_SRAG, SRAG_in_SRAL.
Qed.
