(******************************************************************************)
(** * A technical lemma about monotonicity of write step   *)
(** * (required for the calculation of the predecessors in write steps)   *)
(******************************************************************************)

Require Import Coq.Logic.ClassicalChoice.

Require Import Omega.
Require Import Hahn.

Require Import Basic.
Require Import Filters.
Require Import Subseq.

Require Import Labels Events.
Require Import SRAL.
Require Import Write_lists Match.

Set Implicit Arguments.
Remove Hints plus_n_O.

Section Wmon.

Definition sub_state (st st': State) := 
  forall p ol (IN: st p ol), 
    exists ol', st' p ol' /\ subseq ol ol'.

Lemma sub_state_mapping (st st': State) (SUB: sub_state st st') :
  exists (F : Tid -> list Opt -> list Opt),
  forall p ol (IN: st p ol), st' p (F p ol) /\ subseq ol (F p ol).
Proof.
apply (choice (fun p fol => 
  forall ol (IN: st p ol), st' p (fol ol) /\ subseq ol (fol ol))).
intro p.
apply (choice (fun ol yol => 
  st p ol -> st' p yol /\ subseq ol yol)).
intro ol.
destruct (classic (st p ol)).
- edestruct SUB; eauto.
- exists nil. contradiction.
Qed.

Lemma sub_state_mapping' (st st': State) (SUB: sub_state st st') :
  exists (F : Tid -> list Opt -> list Opt)
         (F' : Tid -> list Opt -> nat -> nat),
  forall p ol (IN: st p ol), st' p (F p ol) /\ 
         subseq_witness ol (F p ol) (F' p ol).
Proof.
apply sub_state_mapping in SUB; desc.
exists F.
apply (choice (fun p fol => 
  forall ol (IN: st p ol), st' p (F p ol) /\ subseq_witness ol (F p ol) (fol ol))).
intro p.
apply (choice (fun ol yol => 
  st p ol -> st' p (F p ol) /\ subseq_witness ol (F p ol) yol)).
intro ol.
destruct (classic (st p ol)).
- edestruct SUB; desf; eauto.
  apply subseq_mapping in H1; desf.
  unfold subseq_witness; eauto.
- exists (fun _ => 0). contradiction.
Qed.

Lemma write_step_monotonicity
          (st st' st0': State) 
          t x v I 
          (WF_ST': wf_state st')
          (SRAL: SRAL_write_step_helper st st' t x v I)
          (WF_ST0': wf_state st0')
          (SUB : sub_state st0' st') : 
          exists st0 I0, sub_state st0 st /\
            SRAL_write_step_helper st0 st0' t x v I0.
Proof.
red in SUB.
eapply sub_state_mapping' in SUB; desc.

set (I0 := fun p ol0' => mk_list (length ol0') 
           (fun n => nth (F' p ol0' n) (I p (F p ol0')) false)).


assert (SUB_I: forall p ol0' (IN: st0' p ol0'), 
  subseq_witness (I0 p ol0') (I p (F p ol0')) (F' p ol0')).
{ ins.
  specialize (SUB _ _ IN); desf.
  red in SRAL; desf.
  red in I_WF.
  specialize (I_WF _ _ SUB).
  red in I_WF; desf.
  red in SUB0; desf.
  subst I0; ins.
  red; splits; auto; ins; rewrite mk_list_length in H.
  2: by rewrite P_LEN; auto.
  rewrite nth_error_nth' with (d:= false).
  2: by rewrite mk_list_length.
  rewrite nth_error_nth' with (d:= false).
  2: by rewrite P_LEN; auto.
  f_equal.
  rewrite nth_mk_list; desf; omega.
}

set (st0 := fun p ol0 => 
  (exists ol0', st0' p ol0' /\ ol0 = filter_ind ol0' (I0 p ol0')) \/
  (p = t /\ exists p' ol0', st0' p' ol0' /\ ol0 = filter_ind_pre ol0' (I0 p' ol0'))).

assert (SUB': sub_state st0 st).
{ unfold st0, I0 in *.
clear st0 I0.
red; ins; desf.
- specialize (SUB _ _ IN); desf.
  specialize (SUB_I _ _ IN); desf.
  red in SRAL; desf.
  specialize (P _ _ SUB); desf.
  red in I_WF.
  specialize (I_WF _ _ SUB).
  red in I_WF; desf.
  cdes SUB0.
  eexists; splits; eauto.
  eapply subseq_mapping_conv with 
    (f := fun n => filter_ind_map (I p (F p ol0'))
       (F' p ol0' (filter_ind_map_inv (mk_list (length ol0') 
           (fun m => nth (F' p ol0' m) (I p (F p ol0')) false)) n))).
  + ins.
    apply filter_ind_map_monotone_strict.
    * apply MON.
      eapply filter_ind_map_inv_monotone_strict; [| done|edone].
      by rewrite mk_list_length.
    * rewrite P_LEN.
      eapply LEN.
      apply filter_ind_map_inv_length; [|done].
      by rewrite mk_list_length.
    * erewrite <- subseq_witness_nth with (d1:=false); [| apply SUB_I|].
      2: { rewrite mk_list_length.
           apply filter_ind_map_inv_length.
           by rewrite mk_list_length.
           omega. }
      eapply filter_ind_map_inv_nin_P with (l:=ol0').
      by rewrite mk_list_length.
      omega.
  + ins.
    rewrite <- filter_ind_map_prop.
    2: by rewrite P_LEN.
    * rewrite <- NTH.
      rewrite nth_error_nth_len with (d:=default_opt); [|eauto].
      2: { apply filter_ind_map_inv_length; try done.
          by rewrite mk_list_length. }
      rewrite nth_error_nth_len with (d:=default_opt); [|eauto].
      2: { apply filter_ind_map_inv_length; try done.
          by rewrite mk_list_length. }
      rewrite filter_ind_map_inv_prop'; try done.
      by rewrite mk_list_length.
    * rewrite nth_error_nth_len with (d:=false); [|eauto].
      2: { rewrite P_LEN. eapply LEN.
           apply filter_ind_map_inv_length; try done.
           by rewrite mk_list_length. }
      erewrite <- subseq_witness_nth with (d1:=false); [| apply SUB_I|].
      2: { rewrite mk_list_length.
           apply filter_ind_map_inv_length; try done.
           by rewrite mk_list_length. }
      rewrite filter_ind_map_inv_nin_P with (l:=ol0'); try done.
      by rewrite mk_list_length.

- specialize (SUB _ _ IN0); desf.
  specialize (SUB_I _ _ IN0); desf.
  red in SRAL; desf.
  specialize (T _ _ SUB); desf.
  red in I_WF.
  specialize (I_WF _ _ SUB).
  red in I_WF; desf.
  cdes SUB0.
  eexists; splits; eauto.
  rename p' into p.

  remember ((mk_list (length ol0')
              (fun n : nat =>
               nth (F' p ol0' n) (I p (F p ol0')) false))) as PP.
  assert (MIN: forall n (H: n < length (filter_ind_pre ol0' PP)),
     min_of_P (I p (F p ol0')) <= 
     F' p ol0' (filter_ind_pre_map_inv PP n)).
  { ins; erewrite subseq_witness_min_of_P.
    2: edone.
    apply Nat.lt_le_incl.
    eapply MON.
    apply min_lt_filter_ind_pre_map_inv.
    apply P_non_empty_min; eauto.
    destruct (classic (exists m, nth_error PP m = Some true)); auto.
    apply filter_ind_pre_P_empty_min with (l:=ol0') in H0.
    rewrite H0 in H; ins; omega.
    subst PP. by rewrite mk_list_length. }

  subst PP.
  eapply subseq_mapping_conv with 
    (f := fun n => filter_ind_pre_map (I p (F p ol0'))
       (F' p ol0' (filter_ind_pre_map_inv (mk_list (length ol0') 
           (fun m => nth (F' p ol0' m) (I p (F p ol0')) false)) n))).
  + ins.
    apply filter_ind_pre_map_monotone_strict.
    * apply MON.
      eapply filter_ind_pre_map_inv_monotone_strict; [| done|edone].
      by rewrite mk_list_length.
    * rewrite P_LEN. eapply LEN.
      apply filter_ind_pre_map_inv_length; [|done].
      by rewrite mk_list_length.
    * apply MIN. omega.
    * erewrite <- subseq_witness_nth with (d1:=false); [| apply SUB_I|].
      2: { rewrite mk_list_length.
           apply filter_ind_pre_map_inv_length; try omega.
           by rewrite mk_list_length. }
      eapply filter_ind_pre_map_inv_nin_P with (l:=ol0').
      by rewrite mk_list_length.
      omega.
  + ins.
    rewrite <- filter_ind_pre_map_prop; auto.
    * rewrite <- NTH.
      rewrite nth_error_nth_len with (d:=default_opt); [|eauto].
      2: { apply filter_ind_pre_map_inv_length; try done.
          by rewrite mk_list_length. }
      rewrite nth_error_nth_len with (d:=default_opt); [|eauto].
      2: { apply filter_ind_pre_map_inv_length; try done.
          by rewrite mk_list_length. }
      rewrite filter_ind_pre_map_inv_prop'; try done.
      by rewrite mk_list_length.
    * rewrite nth_error_nth_len with (d:=false); [|eauto].
      2: { rewrite P_LEN.
           eapply LEN.
           apply filter_ind_pre_map_inv_length; try done.
           by rewrite mk_list_length. }
      erewrite <- subseq_witness_nth with (d1:=false); [| apply SUB_I|].
      2: { rewrite mk_list_length.
           apply filter_ind_pre_map_inv_length; try done.
           by rewrite mk_list_length. }
      rewrite filter_ind_pre_map_inv_nin_P with (l:=ol0'); try done.
      by rewrite mk_list_length.
}

exists st0, I0.
unfold st0, I0.
splits; try edone.
red in SRAL; desf.
red; ins; splits; eauto.

- ins. 
  rename ol into ol0.
  specialize (SUB' _ _ IN); desc.
  eapply T_LOC; try edone.
  eapply subseq_In; eauto.
- ins. 
  rename ol into ol0.
  specialize (SUB' _ _ IN); desc.
  eapply NU; try edone.
  eapply subseq_In; eauto.
- red; ins. 
  red in I_WF.
  rename ol into ol0'.
  specialize (SUB _ _ IN); desc.
  specialize (subseq_witness_nth SUB0); intro NTH'.
  red in SUB0; desc.
  specialize (I_WF _ _ SUB).
  red in I_WF; desc.
  red; splits.
  + apply length_mk_list.
  + ins.
    rewrite NTH' with (d2:=default_opt).
    * eapply P_TID.
      rewrite nth_mk_list in NTH_IN; desf.
    * erewrite <- mk_list_length.
      eapply nth_neq; desf.
  + ins.
    rewrite NTH' with (d2:=default_opt).
    * eapply P_LOC.
      rewrite nth_mk_list in NTH_IN; desf.
    * erewrite <- mk_list_length.
      eapply nth_neq; desf.
  + ins.
    rewrite NTH' with (d2:=default_opt).
  * eapply P_VAL.
    rewrite nth_mk_list in NTH_IN; desf.
  * erewrite <- mk_list_length.
    eapply nth_neq; desf.
- ins; right; splits; eauto.
Qed.

End Wmon.