(******************************************************************************)
(** * Basic definitions for handling "write lists"                       *)
(** * useful for representing lists as functions from Nat + list size    *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Set Implicit Arguments.
Remove Hints plus_n_O.

Section WL.

Variable A : Type.

Definition in_wl (wl : nat -> A) N e : Prop :=
  exists n, n < N /\ wl n = e.

Definition order_wl (wl : nat -> A) N e1 e2 : Prop :=
  exists n1 n2, n1 < n2 /\ n2 < N /\
                wl n1 = e1 /\ wl n2 = e2. 

Definition extend_wl (wl : nat -> A) (e : A) :=
 fun n => if (Nat.eqb 0 n) then e else wl (pred n).

Lemma extend_wl_wl wl :
  extend_wl (fun n => wl (S n)) (wl 0) = wl.
Proof.
apply functional_extensionality; ins.
unfold extend_wl; desf.
apply beq_nat_true in Heq; desf.
apply beq_nat_false in Heq; desf.
rewrite Nat.succ_pred; auto.
Qed.

Lemma extend_wl_in_wl wl e n :
  in_wl (extend_wl wl e) (S n) e.
Proof.
red; exists 0; splits; auto; omega.
Qed.

Lemma in_wl_monotone wl n :
  in_wl (fun m => wl (S m)) n ⊆₁ in_wl wl (S n).
Proof.
unfold in_wl; unfolder; ins; desf.
exists (S n0); splits; try done.
omega.
Qed.

Lemma order_wl_monotone wl n :
  order_wl (fun m => wl (S m)) n ⊆ order_wl wl (S n).
Proof.
unfold order_wl; unfolder; ins; desf.
exists (S n1), (S n2); splits; try done; omega.
Qed.

Lemma order_wlE wl N :
  order_wl wl N ≡
  <| in_wl wl N |> ;; order_wl wl N ;; <| in_wl wl N |>.
Proof.
unfold order_wl, in_wl; split; unfolder; ins; desc; subst; eauto 8.
assert (n1 < N) by omega.
splits; desf; eauto 8.
Qed.

Lemma in_wl_extend_wl wl e n :
  in_wl (extend_wl wl e) (S n) ≡₁ in_wl wl n ∪₁ (eq e).
Proof.
unfold extend_wl, in_wl.
unfolder; split; ins; desf; eauto.
- left. exists n1; splits.
omega.
by rewrite <- pred_Sn.
- exists (S n0); splits.
omega.
eauto.
- exists 0; splits.
omega.
eauto.
Qed.

Lemma order_wl_extend_wl wl e n :
  order_wl (extend_wl wl e) (S n) ≡ 
  order_wl wl n 
  ∪ (eq e) × (in_wl wl n).
Proof.
unfold order_wl, extend_wl, in_wl.
unfolder; split; ins; desf; eauto; try omega;
try eauto 15 using lt_S_n.
exists (S n1), (S n2).
rewrite <- !pred_Sn.
splits; eauto; omega.
exists 0, (S n0); splits; eauto.
omega.
omega.
Qed.

End WL.
