(******************************************************************************)
(** * SRAL simulates SRAG: write step   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import Basic.
Require Import Filters.

Require Import Labels Events Execution.
Require Import SRAL Wmax Add_max SRAG.
Require Import Write_lists Match.

Require Import FG.

Set Implicit Arguments.
Remove Hints plus_n_O.

Lemma LsimG_write (st': State) (G G': execution) 
          t x v (SRAG: SRAG_step G G' (Some (t, (Astore x v))))
          (MATCH': SGmatch st' G') : 
          exists st, SRAL_step st st' (Some (t, (Astore x v))) /\ 
                     SGmatch st G.
Proof.
inversion SRAG; subst; clear SRAG; desf; ins.
specialize (MAX hahn__true_is_true).
clear CO.
assert (MATCH'':=MATCH').
apply SGmatch_SGmatch_fun in MATCH'; destruct MATCH' as [fwl' MATCH'].
rename x0 into t.
rename n into x0.
set (I := fun p ol' => mk_list (length ol') (fun n => 
          if excluded_middle_informative 
           (fwl' p ol' n = (ThreadEvent t x0 (Astore x v))) then true
          else false)).

set (st := fun p ol => 
  (exists ol', st' p ol' /\ ol = filter_ind ol' (I p ol')) \/
  (p = t /\ 
  exists p' ol', st' p' ol' /\ 
(*                  min_of_P (I p ol') <= length (I p ol') /\ *)
                 ol = filter_ind_pre ol' (I p' ol'))).

assert (LEN: forall p ol, length (I p ol) = length ol).
by ins; subst I; ins; rewrite length_mk_list; eauto.

assert (WL_W : forall p ol' (IN: st' p ol'),
  in_wl (fun n : nat => (fwl' p ol') (filter_ind_map_inv (I p ol') n))
  (length (filter_ind ol' (I p ol'))) ⊆₁ W).
{ ins.
  specialize (MATCH' p ol' IN); desc.
  red in MATCH'; desc.
  unfold in_wl; unfolder; ins; desf; splits; ins; desc.
  apply WL_W. unfold in_wl.
  eexists; splits; eauto.
  by eapply filter_ind_map_inv_length.
}

assert (WL_G : forall p ol' (IN: st' p ol'),
  in_wl (fun n : nat => (fwl' p ol') (filter_ind_map_inv (I p ol') n))
  (length (filter_ind ol' (I p ol'))) ⊆₁ acts_set G).
{ ins.
  specialize (MATCH' p ol' IN); desc.
  red in MATCH'; desc.
  rewrite (add_acts_set ADD) in WL_G.
  unfold in_wl; unfolder; ins; desf; splits; ins; desc.
  assert (filter_ind_map_inv (I p ol') n < length ol').
  by eapply filter_ind_map_inv_length; try done.
  specialize (WL_G (fwl' p ol' (filter_ind_map_inv (I p ol') n))).
  exploit WL_G.
  unfold in_wl.
  eexists; splits; eauto.
  unfolder; ins; desf.
  exfalso.
  exploit (filter_ind_map_inv_nin_P ol' (I p ol')); eauto.
  ins.
  subst I; ins; rewrite nth_mk_list in x3.
  rewrite <- x1 in x3.
  desf; try omega.
}

assert (WL_LOC : forall p ol' (IN: st' p ol') n,
        n < length (filter_ind ol' (I p ol')) ->
        loc ((fwl' p ol') (filter_ind_map_inv (I p ol') n)) = 
          loc_opt (nth n (filter_ind ol' (I p ol')) default_opt)).
{ ins.
  specialize (MATCH' p ol' IN); desc.
  red in MATCH'; desc.
  erewrite WL_LOC; try done.
  by rewrite filter_ind_map_inv_prop'.
  by apply filter_ind_map_inv_length.
}

assert (WL_W_PRE : forall p ol' (IN: st' p ol'),
  in_wl (fun n : nat => (fwl' p ol') (filter_ind_pre_map_inv (I p ol') n))
  (length (filter_ind_pre ol' (I p ol'))) ⊆₁ W).
{ ins.
  specialize (MATCH' p ol' IN); desc.
  red in MATCH'; desc.
  unfold in_wl; unfolder; ins; desf; splits; ins; desc.
  apply WL_W0. unfold in_wl.
  eexists; splits; eauto.
  apply filter_ind_pre_map_inv_length; try done.
}

assert (WL_G_PRE : forall p ol' (IN: st' p ol'),
  in_wl (fun n : nat => (fwl' p ol') (filter_ind_pre_map_inv (I p ol') n))
  (length (filter_ind_pre ol' (I p ol'))) ⊆₁ acts_set G).
{ ins.
  specialize (MATCH' p ol' IN); desc.
  red in MATCH'; desc.
  rewrite (add_acts_set ADD) in WL_G0.
  unfold in_wl; unfolder; ins; desf; splits; ins; desc.
  assert (filter_ind_pre_map_inv (I p ol') n < length ol').
  by eapply filter_ind_pre_map_inv_length; try done.
  specialize (WL_G0 (fwl' p ol' (filter_ind_pre_map_inv (I p ol') n))).
  exploit WL_G0.
  unfold in_wl.
  eexists; splits; eauto.
  unfolder; ins; desf.
  exfalso.
  exploit (filter_ind_pre_map_inv_nin_P ol' (I p ol')); eauto.
  ins.
  subst I; ins; rewrite nth_mk_list in x3.
  rewrite <- x1 in x3.
  desf; omega.
}

assert (WL_LOC_PRE : forall p ol' (IN: st' p ol') n,
        n < length (filter_ind_pre ol' (I p ol')) ->
         loc ((fwl' p ol') (filter_ind_pre_map_inv (I p ol') n)) = 
          loc_opt (nth n (filter_ind_pre ol' (I p ol')) default_opt)).
{ ins.
  specialize (MATCH' p ol' IN); desc.
  red in MATCH'; desc.
  erewrite WL_LOC0; try done.
  by rewrite filter_ind_pre_map_inv_prop'.
  by eapply filter_ind_pre_map_inv_length.
}

exists st; splits.
- apply step_write with (I:=I); red; splits; ins.
  + subst st; ins; desf.
    * specialize (MATCH' t ol' IN); desc.
      red in MATCH'; desc.
      red in MATCH'0; desc.
      rewrite add_max_non_read_dom_co_hb in WL_D; try edone.
      2: intros; type_solver.
      eapply In_nth in INo; desc; rewrite <- INo0. 
      rewrite <- WL_LOC; try edone.
      unfold loc_opt; ins.
      intro.
      eapply WL_D.
      split; cycle 1.
      -- right; unfolder; splits; eauto.
         eapply WL_G; try edone.
         by unfold in_wl; eauto 21.
         eapply WL_W; try edone.
         by unfold in_wl; eauto 21.
      -- unfold in_wl.
         eexists; splits; eauto.
         by apply filter_ind_map_inv_length.
    * specialize (MATCH' p' ol' IN0); desc.
      red in MATCH'; desc.
      red in MATCH'0; desc.
      eapply In_nth in INo; desc; rewrite <- INo0. 
      destruct (classic 
       (exists m, nth_error (I p' ol') m = Some true)) as [E|X]; cycle 1.
      by apply filter_ind_pre_P_empty_min with (l:=ol') in X; try done;
        rewrite X in INo; ins; omega.
      rewrite add_max_non_read_co_hb in WL_O; try edone.
      2: intros; type_solver.
      rewrite <- WL_LOC_PRE; try edone.
      unfold loc_opt; ins.
      intro.
      eapply (WL_O (ThreadEvent t x0 (Astore x v))).
      exists (fwl' p' ol' (filter_ind_pre_map_inv (I p' ol') n)); splits.
      -- exists (min_of_P (I p' ol')), (filter_ind_pre_map_inv (I p' ol') n).
         splits; try done.
         by apply min_lt_filter_ind_pre_map_inv.
         by apply filter_ind_pre_map_inv_length.
         exploit (true_in_min (I p' ol')).
         apply P_non_empty_min; try done.
         intro B.
         subst I; ins.
         rewrite mk_list_nth in B; desf.
      -- right.
         unfolder; splits; try done.
         right; splits; try done.
         by eapply WL_G_PRE; try edone; red; eauto.
         by eapply WL_W_PRE; try edone; red; eauto.
  + subst st; ins; desc; intro H.
eapply In_nth in INo; desc. 
      rewrite <- INo0 in *; clear INo0.
    desf.
    * specialize (MATCH' p ol' IN); desc.
      red in MATCH'; desc.
      red in MATCH'0; desc.
      eapply WL_RMW; try edone.
      eapply filter_ind_map_inv_length; try edone.
      rewrite filter_ind_map_inv_prop'; try edone.
      exists (ThreadEvent t x0 (Astore x v)).
      eapply add_max_write_co; try edone.
      by type_solver.
      right; unfolder; splits; try done.
      eapply WL_G; try edone; unfold in_wl; eauto.
      eapply WL_W; try edone; unfold in_wl; eauto.
      by unfold is_loc; rewrite WL_LOC.
    * specialize (MATCH' p' ol' IN0); desc.
      red in MATCH'; desc.
      red in MATCH'0; desc.
      eapply WL_RMW; try edone.
      eapply filter_ind_pre_map_inv_length; try edone.
      rewrite filter_ind_pre_map_inv_prop'; try edone.
      exists (ThreadEvent t x0 (Astore x v)).
      eapply add_max_write_co; try edone.
      by type_solver.
      right; unfolder; splits; try done.
      eapply WL_G_PRE; try edone; unfold in_wl; eauto.
      eapply WL_W_PRE; try edone; unfold in_wl; eauto.
      by unfold is_loc; rewrite WL_LOC_PRE.
  + red; ins; red; splits; ins.
    * assert (N: n < length ol).
      by erewrite <- LEN; eapply nth_neq; intro; desf.
      subst I; ins; rewrite nth_mk_list in NTH_IN; desf.
      specialize (MATCH' p ol IN); desc.
      red in MATCH'; desc.
      rewrite <- WL_TID; try done.
      red in ADD; desc; ins.
      rewrite e; ins.
    * assert (N: n < length ol).
      by erewrite <- LEN; eapply nth_neq; intro; desf.
      subst I; ins; rewrite nth_mk_list in NTH_IN; desf.
      specialize (MATCH' p ol IN); desc.
      red in MATCH'; desc.
      rewrite <- WL_LOC0; try done.
      red in ADD; desc; ins.
      rewrite e; ins.
    * assert (N: n < length ol).
      by erewrite <- LEN; eapply nth_neq; intro; desf.
      subst I; ins; rewrite nth_mk_list in NTH_IN; desf.
      specialize (MATCH' p ol IN); desc.
      red in MATCH'; desc.
      rewrite <- WL_VAL; try done.
      red in ADD; desc; ins.
      rewrite e; ins.
  + by subst st; ins; desf; eauto.
  + by right; splits; eauto.
  + apply MATCH''.
- red; splits; try done.
  { red in MATCH''; desc; red in WF_ST.
    red; subst st; ins.
    specialize (WF_ST t0); desc; eauto 12.
  }
  ins; subst st; desf; ins; desf; desc.
  + specialize (MATCH' p ol' IN); desc.
    red in MATCH'; desc.
    rewrite (add_acts_set ADD) in WL_G0.
    red in MATCH'0; desc.
    rewrite add_max_non_read_co_hb in WL_O; try edone.
    2: intros; type_solver.
    rewrite add_max_non_read_dom_co_hb in WL_D; try edone.
    2: intros; type_solver.
    exists (fun n => (fwl' p ol') (filter_ind_map_inv (I p ol') n)).
    red; splits; red; splits; eauto.
    * ins.
      erewrite WL_TID; try done.
      by rewrite filter_ind_map_inv_prop'.
      by apply filter_ind_map_inv_length.
    * ins.
      erewrite WL_VAL; try done.
      by rewrite filter_ind_map_inv_prop'.
      by apply filter_ind_map_inv_length.
    * ins.
      intro A; eapply WL_RMW.
      eby apply filter_ind_map_inv_length.
      by rewrite filter_ind_map_inv_prop'.
      revert A; generalize (add_co_monotone ADD); basic_solver 21.
    * unfold in_wl; unfolder; ins; desc; splits; ins; desc.
      eapply WL_D.
      split.
      2: by unfolder; eauto 12.
      exists (filter_ind_map_inv (I p ol') n); splits; try done.
      by apply filter_ind_map_inv_length.
    * unfold in_wl; unfolder; ins; desc; splits; ins; desc.
      eapply WL_O.
      unfold in_wl; unfolder; splits.
      eexists; splits; eauto.
      unfold order_wl in *; desc.
      eexists (filter_ind_map_inv (I p ol') n1).
      eexists (filter_ind_map_inv (I p ol') n2).
      splits; eauto.
      eapply filter_ind_map_inv_monotone_strict with (l:=ol'); eauto.
      by apply filter_ind_map_inv_length.
  + specialize (MATCH' p' ol' IN0); desc.
    red in MATCH'; desc.
    rewrite (add_acts_set ADD) in WL_G0.
    red in MATCH'0; desc.
    rewrite add_max_non_read_co_hb in WL_O; try edone.
    2: intros; type_solver.
    rewrite add_max_non_read_dom_co_hb in WL_D; try edone.
    2: intros; type_solver.
    exists (fun n => (fwl' p' ol') (filter_ind_pre_map_inv (I p' ol') n)).
    red; splits; red; splits; eauto.
    * ins.
      erewrite WL_TID; try done.
      by rewrite filter_ind_pre_map_inv_prop'.
      by apply filter_ind_pre_map_inv_length.
    * ins.
      erewrite WL_VAL; try done.
      by rewrite filter_ind_pre_map_inv_prop'.
      by apply filter_ind_pre_map_inv_length.
    * ins.
      intro A; eapply WL_RMW.
      eby apply filter_ind_pre_map_inv_length.
      by rewrite filter_ind_pre_map_inv_prop'.
      revert A; generalize (add_co_monotone ADD); basic_solver 21.
    * unfold in_wl; unfolder; ins; desc; splits; ins; desc.
      destruct (classic 
       (exists m, nth_error (I p' ol') m = Some true)) as [E|X]; cycle 1.
      by apply filter_ind_pre_P_empty_min with (l:=ol') in X; try done;
        rewrite X in H; ins; omega.
      eapply (WL_O (ThreadEvent t x0 (Astore x v))). 
      exists x1; splits; [|basic_solver 21].
      exists (min_of_P (I p' ol')), (filter_ind_pre_map_inv (I p' ol') n).
      splits; try done.
      by apply min_lt_filter_ind_pre_map_inv.
      by apply filter_ind_pre_map_inv_length.
      exploit (true_in_min (I p' ol')).
      apply P_non_empty_min; try done.
      intro B.
      subst I; ins.
      rewrite mk_list_nth in B; desf.
    * unfold in_wl; unfolder; ins; desc; splits; ins; desc.
      eapply WL_O.
      unfold in_wl; unfolder; splits.
      eexists; splits; eauto.
      unfold order_wl in *; desc.
      eexists (filter_ind_pre_map_inv (I p' ol') n1).
      eexists (filter_ind_pre_map_inv (I p' ol') n2).
      splits; eauto.
      eapply filter_ind_pre_map_inv_monotone_strict with (l:=ol'); eauto.
      by apply filter_ind_pre_map_inv_length.
Qed.
